#include "Panel_GAMESS.h"

#include "Log.h"

#include "Defs.h"
#include "Path.h"
#include "User.h"
#include "Model.h"
#include "Model_kit.h"
#include "Atom.h"
#include "Main_window.h"

#include <QCloseEvent>
#include <QDir>

#include <sstream>
#include <fstream>
#include <math.h>

namespace MM
{
using namespace std;

Panel_GAMESS::
Panel_GAMESS ()
:
    //Panel ("GAMESS", "left"),
    Panel ("Firefly", "left"),

    model_                  (0),
    process_                (new QProcess(this)),
    //output_                 (Path::GAMESS() + "output"),
    output_                 (setup_gamess_dir() + "output"),
    player_                 (Main_window::singleton()),
    read_atom_coords_flag_  (false),
    read_charges_flag_      (false),
    read_moments_flag_      (false),
    user_stop_flag_         (false),
    success_                (false),
    optimization_it_        (0)
{
    setupUi (this);

    //====================================================

    //for (int i=5;  i<listWidget_Basis->count();  ++i)
    //    listWidget_Basis->item(i)->setHidden (true);

    listWidget_Basis->item(5)->setHidden (true);

    //====================================================

    connect (tabWidget_Basis,   SIGNAL(currentChanged(int)),
             this,              SLOT  (basis_panel   ()));


    connect (comboBox_RUNTYP,   SIGNAL(currentIndexChanged (const QString &)),
             this,              SLOT  (show_options        (const QString &)));

    process_->setReadChannel (QProcess::StandardOutput);
    process_->setWorkingDirectory (gamess_dir());


    connect (pushButton_Check,  SIGNAL(clicked()),
             this,              SLOT  (check  ()));



    connect (process_,          SIGNAL(started()),
             this,              SLOT  (started()));

    connect (process_,          SIGNAL(finished (int, QProcess::ExitStatus)),
             this,              SLOT  (finished (int, QProcess::ExitStatus)));

    connect (process_,          SIGNAL(readyRead()),
             this,              SLOT  (read()));

   
    listWidget_Basis->setCurrentRow (3);



    //connect (checkBox_BSSE,             SIGNAL(toggled       (bool)),
    //         this,                      SLOT  (set_BSSE_calc (bool)));

    connect (doubleSpinBox_with_BSSE_y,     SIGNAL(valueChanged (double)),
             this,                          SLOT  (extrapolate ()));
    connect (doubleSpinBox_with_BSSE_x,     SIGNAL(valueChanged (double)),
             this,                          SLOT  (extrapolate ()));
    //connect (doubleSpinBox_without_BSSE_y,  SIGNAL(valueChanged (double)),
    //         this,                          SLOT  (extrapolate ()));
    //connect (doubleSpinBox_without_BSSE_x,  SIGNAL(valueChanged (double)),
    //         this,                          SLOT  (extrapolate ()));
    connect (doubleSpinBox_Alpha,           SIGNAL(valueChanged (double)),
             this,                          SLOT  (extrapolate ()));
    connect (doubleSpinBox_Y,               SIGNAL(valueChanged (double)),
             this,                          SLOT  (extrapolate ()));
    connect (doubleSpinBox_X,               SIGNAL(valueChanged (double)),
             this,                          SLOT  (extrapolate ()));

}

Panel_GAMESS::
~Panel_GAMESS ()
{
    if (process_->isOpen ())
        process_->kill ();

    if (output_.isOpen ())
    {
        output_.flush ();
        output_.close ();
    }
}

void Panel_GAMESS::
turn (bool on)
{
    //Main_window *player = Main_window::singleton();

    //if (process_->state() == QProcess::NotRunning)
    if (!player_->player_on_)
    {
        player_->toolButton_Stop   -> disconnect ();
        player_->toolButton_Back   -> disconnect ();
        player_->toolButton_Tune   -> disconnect ();
        player_->toolButton_Run    -> disconnect ();
        player_->toolButton_Start  -> disconnect ();
        player_->toolButton_Forward-> disconnect ();
        player_->toolButton_Pause  -> disconnect ();

        if (on)
        {
            player_->tabWidget_bottom->setCurrentWidget (player_->tab_Run);

            connect (player_->toolButton_Stop,   SIGNAL(clicked()), this,
                                 SLOT  (stop   ()));
            connect (player_->toolButton_Tune,   SIGNAL(clicked()), this,
                                 SLOT  (tune   ()));
            connect (player_->toolButton_Run,    SIGNAL(clicked()), this,
                                 SLOT  (run    ()));
            connect (player_->toolButton_Start,  SIGNAL(clicked()), this,
                                 SLOT  (start  ()));
            set_stop_state ();
        }
        else 
        {
            player_->toolButton_Stop   -> setEnabled (false);
            player_->toolButton_Back   -> setEnabled (false);
            player_->toolButton_Tune   -> setEnabled (false);
            player_->toolButton_Run    -> setEnabled (false);
            player_->toolButton_Start  -> setEnabled (false);
            player_->toolButton_Forward-> setEnabled (false);
            player_->toolButton_Pause  -> setEnabled (false);
        }
    }
}

void Panel_GAMESS::
closeEvent (QCloseEvent * event)
{
    if (process_->state() == QProcess::Running)
        event->ignore ();
}

Text const & Panel_GAMESS::
setup_gamess_dir ()
{
    //gamess_dir_ = Path::GAMESS();
    gamess_dir_ = Path::Firefly();

    ifstream path_file ((gamess_dir_ + "path.txt").c_str());

    if (path_file)
    {
        string new_path;
        getline (path_file, new_path);
        gamess_dir_ = new_path.c_str();
    }

    return gamess_dir_;
}

void Panel_GAMESS::
show_options (const QString & text)
{
    for (int i=0;  i<stackedWidget->count();  ++i)
    {
        if (stackedWidget->widget(i)->objectName() == text)
            stackedWidget->setCurrentIndex (i);
    }
}

void Panel_GAMESS::
basis_panel ()
{
    Text panel = tabWidget_Basis->currentWidget()->objectName();

    if (panel == "Options")
    {
        pushButton_Check->setEnabled (false);
        turn (false);
    }
    else
    {
        pushButton_Check->setEnabled (true);
        turn (true);
    }
}

void Panel_GAMESS::
run ()
{
    if (process_->state() == QProcess::NotRunning)
    {
        tune  ();
        start ();
    }
    else
        stop();
}

void Panel_GAMESS::
check ()
{
    model_ = &model();

    if (pushButton_Check->isChecked())
    {
        create_input  (" $CONTRL EXETYP=CHECK   $END");
        start ();
        //to_user().status ("GAMESS checking started.");
    }
    else
        stop();
}

void Panel_GAMESS::
tune ()
{
    model_ = &model();

    create_input (" $CONTRL EXETYP=RUN      $END");
    to_user().status ("GAMESS input generated.");
}

void Panel_GAMESS::
start ()
{
    model_ = &model();

    if (process_->state() == QProcess::NotRunning)
    {
        to_user().info ("\n___________________________________________");

        if (output_.isOpen ())
            output_.close ();

        if (!output_.open (QIODevice::WriteOnly/* | QIODevice::Text*/))
        {
            to_user().error ("Can not open 'output' file.");     
            return;
        }

        QString name = stackedWidget->currentWidget()->objectName();//fix to state
        if (name == "OPTIMIZE")
        {
            optimization_it_ = 1;
        }
        else
            optimization_it_ = 0;

        QDir dir(gamess_dir());
        dir.remove ("IRCDATA");
        dir.remove ("PUNCH");
        dir.remove ("AOINTS");
        dir.remove ("DICTNRY");

        // Start ==============================================
        Text        program = gamess_dir() +"firefly";
        QStringList arguments;
        arguments << "-f";
        //log () << program;

        user_stop_flag_ = false;

        if (external_basis_.is_not_empty())
        {
            arguments << "-b" << (external_basis_ + ".lib").c_str();
            process_->start (program, arguments);
        }
        else
            process_->start (program);

        bool started = process_->waitForStarted (3000);
        // Start ==============================================

        if (!started)
        {

            //Text path = Path::GAMESS();
	    //log () << path;

            //ifstream path_file ((path + "path.txt").c_str());
            //if (path_file)
            //{
            //    string new_path;
            //    getline (path_file, new_path);
            //    path = new_path.c_str();
            //}

            //Text        program = Path::GAMESS() +"pcgamess.exe";
            Text        program = gamess_dir() +"pcgamess";
            QStringList arguments;
            arguments << "-f";
    //	log () << program;

            user_stop_flag_ = false;

            if (external_basis_.is_not_empty())
            {
                arguments << "-b" << (external_basis_ + ".lib").c_str();
                process_->start (program, arguments);
            }
            else
                process_->start (program);

            started = process_->waitForStarted (3000);
        }

        if (!started)
        {
            to_user().fatal_error (
                "Can not start Firefly (PC GAMESS).\n\n"
                "Please check that there are \"Firefly\" and dynamic library "
                "files in Ascalaph/Firefly folder or in the folder written "
                "in Ascalaph/Firefly/path.txt file.\n"
                "\n"
                "Firefly is not a part of Ascalaph.\n"
                "If you don't have it, take it for free from "
                "http://classic.chem.msu.su/gran/gamess/index.html"
                );
            set_stop_state ();
        }
    }
    else
        stop();
}

void Panel_GAMESS::
stop ()
{
//    pushButton_Check->setEnabled (false);
    user_stop_flag_ = true;
    player_->player_on_ = false;
    process_->kill();

    //to_user ().info ("Simulation was stopped by the user. "
    //    "Please wait for the end of trajectory reading.");
}

void Panel_GAMESS::
started ()
{
    //to_user ().status ("GAMESS started");
    to_user ().status ("Firefly started");
    set_run_state ();
}

void Panel_GAMESS::
finished (int, QProcess::ExitStatus exitStatus)
{
    //qApp->processEvents(QEventLoop::ExcludeSocketNotifiers);

    if (exitStatus == QProcess::CrashExit && !user_stop_flag_)
    {
        //Text message ("GAMESS crashed. ");
        Text message ("Firefly crashed. ");
        to_user ().status (message);

        switch (process_->error ())
        {
        case QProcess::FailedToStart:
            to_user ().error (message +
                "The process failed to start. "
                "Either the invoked program is missing, or you may have "
                "insufficient permissions to invoke the program.");
            break;

        case QProcess::Crashed:
            to_user ().error (message +
                "The process crashed some time after "
                "starting successfully.");
            break;

        case QProcess::Timedout:
            to_user ().error (message + "Timed out");
            break;

        case QProcess::WriteError:
            to_user ().error (message +
                "An error occurred when attempting to write to the process.");
            break;

        case QProcess::ReadError:
            to_user ().error (message +
                "An error occurred when attempting to read from "
                "the process. For example, the process may not be running.");
            break;

        case QProcess::UnknownError:
            to_user ().error (message + "An unknown error occurred.");
            break;
        }
    }
    else
    {
        //read ();
        output_.flush ();

        //Text message ("GAMESS finished ");
        Text message ("Firefly finished ");
        if (success_)
            message += "NORMALLY";
        //else
        //    to_user ().warning ("Execution of GAMESS terminated ABNORMALLY.");

        to_user ().status (message);
    }

    read ();
    set_stop_state ();
}

void Panel_GAMESS::
set_stop_state ()
{
    player_->player_on_ = false;

    pushButton_Check->setChecked (false);
    pushButton_Check->setEnabled (true);

    player_->toolButton_Stop   -> setEnabled (false);
    player_->toolButton_Tune   -> setEnabled (true);
    player_->toolButton_Run    -> setEnabled (true);
    player_->toolButton_Start  -> setEnabled (true);

    if (this->isHidden ())
        turn (false);
}

void Panel_GAMESS::
set_run_state ()
{
    player_->player_on_ = true;

    pushButton_Check->setEnabled (false);
    if (pushButton_Check->isChecked())
        pushButton_Check->setEnabled(true);

    player_->toolButton_Stop   -> setEnabled (true);
    player_->toolButton_Tune   -> setEnabled (false);
    player_->toolButton_Run    -> setEnabled (false);
    player_->toolButton_Start  -> setEnabled (false);
}

void Panel_GAMESS::
read ()
{
    char    buf[1024];
    //Text    text;
    char *  result = 0;
    string  stub;
    //Text    message;

    while (process_->canReadLine ())
    {
        qint64 bytes = process_->readLine (buf, sizeof(buf));
        output_.write (buf, bytes);

        for (int i=0;  i<1024;  ++i)
        {
            if (buf[i] == '\0')
            {
                buf[i-2] = 0x0A;
                buf[i-2] = '\0';
            }
        }

        if (read_atom_coords_flag_)
        {
            read_atom_coords (buf);
            continue;
        }

        if (read_charges_flag_)
        {
            read_charges (buf);
            continue;
        }

        if (read_moments_flag_)
        {
            read_moments (buf);
            continue;
        }

        //=====================================================================
        qApp->processEvents (QEventLoop::ExcludeSocketNotifiers);

        //result = strstr (buf, "1NSERCH=");
        //if (result)
        //    sscanf (buf, "1NSERCH=%i", & NSERCH_ );

        result = strstr (buf, "EQUILIBRIUM GEOMETRY LOCATED");
        if (result)
        {
            to_user ().info (buf);
            continue;
        }

        result = strstr (buf, "COORDINATES OF ALL ATOMS ARE (ANGS)");
        if (result || 
            strstr (buf, "COORDINATES OF SYMMETRY UNIQUE ATOMS (ANGS)"))
        {
            read_atom_coords_flag_ = true;
            continue;
        }

        //result = strstr (buf, "ELECTROSTATIC MOMENTS");
        //if (result)
        //{
        //    to_user().info ("");
        //    read_moments_flag_ = true;
        //    read_moments (buf);
        //}

        result = strstr (buf, "NET CHARGES:");
        if (result)
        {
            to_user().info ("");
            read_charges_flag_ = true;
            read_charges (buf);
            continue;
        }

        result = strstr (buf, "MULTIPOLE MOMENTS:");
        if (result)
        {
            to_user().info ("");
            read_moments_flag_ = true;
            read_moments (buf);
            continue;
        }

        double e;

        result = strstr (buf, "FINAL ENERGY IS");
        if (result)
        {
            if (optimization_it_!=0)
                continue;

            istringstream in (buf);
            int    it;
            in >> stub >> stub >> stub >> e >> stub >> it;

            Text message;
            message += Text (" E        = ");
            message += Text (e, 15);
            message += " Hartree   ";
            message += Text (e*Constant::Hartree_to_kcal, 15);
            message += " kcal/mol   after ";
            message += it;
            message += " iterations.";
            to_user ().info (message);
            message.clear ();


            if (radioButton_Ex->isChecked())
                doubleSpinBox_with_BSSE_x->setValue (e);
            if (radioButton_Ey->isChecked())
                doubleSpinBox_with_BSSE_y->setValue (e);

            continue;
        }

        result = strstr (buf, "E(MP2)=");
        if (result)
        {
            if (optimization_it_!=0)
                continue;

            istringstream in (buf);     double e;   in >> stub >> e;
            Text message ("E(MP2) = "); 
            message += Text (e, 15); message += " Hartree   "; 
            message += Text (e*Constant::Hartree_to_kcal, 15); message += " kcal/mol";
            to_user ().info (message);
            //message += Text (" E(MP2) = "); message += Text (e, 15);

            //if (!optimization_it_!=0)
            //{
            //    to_user ().info (message);
            //    message.clear ();
            //}

            if (radioButton_Ex->isChecked())
                doubleSpinBox_with_BSSE_x->setValue (e);
            if (radioButton_Ey->isChecked())
                doubleSpinBox_with_BSSE_y->setValue (e);

            continue;
        }

        result = strstr (buf, "E(MP2)");
        if (result)
        {
            if (optimization_it_!=0)
                continue;

            istringstream in (buf);     double e;   in >> stub >> stub >> e;
            Text message ("E(MP2) = "); 
            message += Text (e, 15); message += " Hartree   "; 
            message += Text (e*Constant::Hartree_to_kcal, 15); message += " kcal/mol";
            to_user ().info (message);
            //message += Text (" E(MP2) = "); message += Text (e, 15);

            //if (!optimization_it_!=0)
            //{
            //    to_user ().info (message);
            //    message.clear ();
            //}

            if (radioButton_Ex->isChecked())
                doubleSpinBox_with_BSSE_x->setValue (e);
            if (radioButton_Ey->isChecked())
                doubleSpinBox_with_BSSE_y->setValue (e);

            continue;
        }

        result = strstr (buf, "E(MP3)");
        if (result)
        {
            istringstream in (buf);     double e;   in >> stub >> stub >> e;
            Text message ("E(MP3) = "); 
            message += Text (e, 15); message += " Hartree   "; 
            message += Text (e*Constant::Hartree_to_kcal, 15); message += " kcal/mol";
            //message += Text (" E(MP3) = "); message += Text (e, 15);
            to_user ().info (message);

            if (radioButton_Ex->isChecked())
                doubleSpinBox_with_BSSE_x->setValue (e);
            if (radioButton_Ey->isChecked())
                doubleSpinBox_with_BSSE_y->setValue (e);

            continue;
        }

        result = strstr (buf, "E(MP4");
        if (result)
        {
            istringstream in (buf);     double e;   in >> stub >> stub >> e;
            Text message ("E(MP4) = "); 
            message += Text (e, 15); message += " Hartree   "; 
            message += Text (e*Constant::Hartree_to_kcal, 15); message += " kcal/mol";
            //message += Text (" E(MP4) = "); message += Text (e, 15);
            to_user ().info (message);

            if (radioButton_Ex->isChecked())
                doubleSpinBox_with_BSSE_x->setValue (e);
            if (radioButton_Ey->isChecked())
                doubleSpinBox_with_BSSE_y->setValue (e);

            continue;
        }

        result = strstr (buf, "MAXIMUM GRADIENT");
        if (result)
        {
            istringstream in (buf);
            double max, rms;
            in >> stub >> stub >> stub >> max >> stub >> stub >> stub >> rms;

            Text message ("Gradient ");

            if (optimization_it_!=0)
            {
                message += optimization_it_;
                message += " ";
            }

            message += Text (" Max ");
            message += max;
            message += "&nbsp;&nbsp;&nbsp; RMS ";
            message += rms;
            to_user ().info (message);
            message.clear ();
            ++optimization_it_;
            continue;
        }

        //if (message.is_not_empty())
        //{
        //    to_user ().info (message);
        //    message.clear ();
        //}

        //=====================================================================
        char *abnormal;

        abnormal= 
            //strstr (buf, "EXECUTION OF GAMESS TERMINATED ABNORMALLY");
            strstr (buf, "EXECUTION OF Firefly TERMINATED ABNORMALLY");
        if (abnormal)
            to_user ().fatal_error (buf);

        abnormal = strstr (buf, "*** ERROR");
        if (abnormal || strstr (buf, "ERROR!"))
            to_user ().error (buf);

        //char *normal = strstr (buf, "EXECUTION OF GAMESS TERMINATED NORMALLY");
        char *normal = strstr (buf, "EXECUTION OF Firefly TERMINATED NORMALLY");
        if (normal)
            success_ = true;
    }
    //log () << text;
    output_.flush ();
//    qApp->processEvents (QEventLoop::ExcludeSocketNotifiers);
}

void Panel_GAMESS::
set_BSSE_calc (bool /*on*/)
{
    //if (on)
    //{
    //    doubleSpinBox_without_BSSE_y->setEnabled (true);
    //    doubleSpinBox_without_BSSE_x->setEnabled (true);
    //    doubleSpinBox_E_without_BSSE->setEnabled (true);
    //}
    //else
    //{
    //    doubleSpinBox_without_BSSE_y->setEnabled (false);
    //    doubleSpinBox_without_BSSE_x->setEnabled (false);
    //    doubleSpinBox_E_without_BSSE->setEnabled (false);
    //}

}

void Panel_GAMESS::
extrapolate ()
{
    double Alpha = doubleSpinBox_Alpha->value();
    double Ex    = doubleSpinBox_with_BSSE_x->value();
    double Ey    = doubleSpinBox_with_BSSE_y->value();
    double X     = doubleSpinBox_X->value();
    double Y     = doubleSpinBox_Y->value();
    X = pow (X, Alpha);
    Y = pow (Y, Alpha);

    double E = (Ex * Y - Ey * X) / (Y - X);
    doubleSpinBox_E->setValue (E);

    //Ex    = doubleSpinBox_without_BSSE_x->value();
    //Ey    = doubleSpinBox_without_BSSE_y->value();

    //E = (Ex * Y - Ey * X) / (Y - X);
    //doubleSpinBox_E_without_BSSE->setValue (E);
}

void Panel_GAMESS::
read_atom_coords (char *buf)
{
    static int n = -2;
    string stub;
    double x, y, z;

    int length = strlen (buf);
    if (length < 4)
    {
        read_atom_coords_flag_ = false;
        n = -2;
        model_->kit().event().conformation_changed ();
        return;
    }

    if (n >=0)
    {
        Atom & atom = model_->atom (n);

        istringstream in (buf);
        in >> stub >> stub >> x >> y >> z;
        atom.set_x (x);
        atom.set_y (y);
        atom.set_z (z);
    }

    ++n;
}

void Panel_GAMESS::
read_charges (char *buf)
{
    static int n = -4;
    string stub;
    double charge;

    if (n >= 0 && buf[1] == '-')
    {
        read_charges_flag_ = false;
        n = -4;
        return;
    }
    to_user().info (buf);

    if (n >=0)
    {
        Atom & atom = model_->atom (n);

        istringstream in (buf);
        in >> stub >> charge;
        atom.set_charge (charge);
    }

    ++n;
}

void Panel_GAMESS::
read_moments (char *buf)
{
    static int n = 0;
    ////string stub;
    ////double dx, dy, dz, moment;

    //int length = strlen (buf);
    //if (n >= 3 && length < 4)
    //{
    //    read_moments_flag_ = false;
    //    n = 0;
    //    return;
    //}

    //to_user().info (buf);

    ////if (n >=0)
    ////{
    ////    Atom & atom = model_->atom (n);

    ////    istringstream in (buf);
    ////    in >> stub >> stub >> x >> y >> z;
    ////    atom.set_x (x);
    ////    atom.set_y (y);
    ////    atom.set_z (z);
    ////}

    if (n > 3 && buf[1] == '-')
    {
        read_moments_flag_ = false;
        n = 0;
        return;
    }
    to_user().info (buf);

    ++n;
}

void Panel_GAMESS::
create_input (char const* first_string)
{
    success_ = false;

    Text text;
    ofstream out ((gamess_dir() + "input").c_str());
    out << first_string << endl;

    text = comboBox_RUNTYP->currentText();
    if (text == "Electrostatics" || text == "Extrapolation")
        text = "ENERGY";
    out << " $CONTRL RUNTYP=" << text.c_str() << "   $END" << endl;
    out << endl;

    out << " $CONTRL ICHARG=" << spinBox_Charge->value()       
        << "        $END"   << endl;
    out << " $CONTRL MULT="   << spinBox_Multiplicity->value() 
        << "          $END" << endl;
    out << " $CONTRL NPRINT=" << spinBox_NPRINT->value() 
        << "       $END" << endl;
    out << " $CONTRL INTTYP=HONDO    $END" << endl;
    out << endl;


    if (not_semiempirical ())
    {
        if (radioButton_MP2->isChecked())
        {
            out << " $CONTRL MPLEVL=2        $END" << endl;
            out << endl;
        }
        else if (radioButton_MP3->isChecked())
        {
            out << " $CONTRL MPLEVL=3        $END" << endl;
            out << endl;
        }
        else if (radioButton_MP4->isChecked())
        {
            out << " $CONTRL MPLEVL=4        $END" << endl;
            out << " $MP4 SDTQ=.true.        $END" << endl;
            out << endl;
        }
        else if (radioButton_DFT->isChecked())
        {
            out << " $CONTRL DFTTYP=";
            out << Text(comboBox_DFT->currentText ()).c_str();
            out <<    "  $END" << endl;
            out << endl;
        }
    }

    if (is_spherical_basis_set())
    {
        out << " $CONTRL D5=.T.                         $END" << endl;
    }

    if (comboBox_Calc->currentText() == "Normal")
    {
        out << " $CONTRL ICUT=11 ITOL=23 INTTYP=HONDO   $END" << endl;
        out << " $SCF soscf=0 diis=1 NCONV=7 MAXDII=15  $END" << endl; 
        out << " $MP2    CPTOL=1e-11                    $END" << endl;
        out << " $MP2    CUTOFF=1e-15                   $END" << endl;
        out << " $MP3    CUTOFF=1e-15                   $END" << endl;
        out << " $MP4    CUTOFF=1e-15                   $END" << endl;
        out << " $MP2GRD TOL1=1D-11 TOL2=1D-11          $END" << endl;

        out << endl;
    }
    else if (comboBox_Calc->currentText() == "Accurate")
    {
        //out << " $CONTRL ICUT=20 ITOL=30 INTTYP=HONDO $END" << endl;
        //out << " $CONTRL ICUT=15 ITOL=30 INTTYP=HONDO $END" << endl;
        out << " $CONTRL ICUT=13 ITOL=25 INTTYP=HONDO   $END" << endl;

        //out << " $SCF    NCONV=15 FDIFF=.FALSE.       $END" << endl;
        out << " $SCF soscf=0 diis=1 NCONV=10 MAXDII=20  $END" << endl; //!!!

        //out << "# $intgrl autocl=.f.                     $END" << endl;

        out << " $MP2    CPTOL=1e-12                    $END" << endl;
        out << " $MP2    CUTOFF=1e-15                   $END" << endl;
        out << " $MP3    CUTOFF=1e-15                   $END" << endl;
        out << " $MP4    CUTOFF=1e-15                   $END" << endl;
        out << " $MP2GRD TOL1=1D-12 TOL2=1D-12          $END" << endl;

        out << endl;
    }

    out << basis().c_str() << endl; //fix max 8 letters !!!

    out << " $CONTRL MAXIT=100       $END" << endl;
    if (spinBox_Multiplicity->value() == 1)
        out << " $CONTRL SCFTYP=RHF      $END" << endl;
    else
        out << " $CONTRL SCFTYP=UHF      $END" << endl;
    out << " $CONTRL NZVAR=0         $END" << endl;
    out << " $CONTRL IREST=0         $END" << endl;
    out << " $CONTRL GEOM=INPUT      $END" << endl;
    out << endl;

    out << " $SCF DEM=.FALSE.        $END" << endl;
    out << " $SCF dirSCF=.TRUE.      $END" << endl;
    out << endl;

    QString name = stackedWidget->currentWidget()->objectName();
    if (name == "OPTIMIZE")
    {
        Text method;
        switch (comboBox_METHOD->currentIndex())
        {
        case 0:     method = "GDIIS noreg=1"; break;
        case 1:     method = "QA";          break;
        case 2:     method = "NR";          break;
        case 3:     method = "RFO";         break;
        case 4:     method = "SCHLEGEL";    break;
        case 5:     method = "CONOPT";      break;
        }
        out << " $STATPT METHOD=" << method.c_str() 
            << "    $END" << endl;

        out << " $STATPT OPTTOL=" << Text(lineEdit_Gradient->text ()).c_str()
            << "    $END" << endl;
        out << " $STATPT NSTEP=" << spinBox_NSTEP->value() 
            << "      $END" << endl;
        out << " $STATPT DXMAX=" << doubleSpinBox_DXMAX->value() 
            << "       $END" << endl;
        out << " $STATPT TRMAX=" << doubleSpinBox_TRMAX->value() 
            << "       $END" << endl;
        out << endl;
    }

    out << " $SYSTEM TIMLIM=6000     $END" << endl;
    out << " $SYSTEM MEMORY=" << Text(lineEdit_MEMORY->text()).c_str()
        << "    $END" << endl;
    out << endl;

    if (! radioButton_No_MP->isChecked())
    {
        out << " $MP2 GRDMET=2           $END" << endl;
        out << " $MP2 DM2MET=-1          $END" << endl;
        out << " $MP2 MNRECL=1000        $END" << endl;
        out << endl;
    }

    if (comboBox_COORD->currentText() == "CART")
    {
        out << " $CONTRL COORD=CART      $END" << endl;
        out << endl;
    }

    name = stackedWidget->currentWidget()->objectName();
    if (name == "Electrostatics")
    {
        out << " $ELPOT IEPOT=1          $END" << endl;
        out << " $ELPOT WHERE=PDC        $END" << endl;
        out << " $PDC   CONSTR="<< Text(comboBox_CONSTR->currentText()).c_str()
            << "    $END" << endl;
        out << " $PDC   NFREQ="  << spinBox_NFREQ->value ()
            << "         $END" << endl;
        out << " $PDC   MAXPDC=" << spinBox_MAXPDC->value ()
            << "        $END" << endl;
        out << " $PDC   VDWSCL=" << doubleSpinBox_VDWSCL->value ()
            << "        $END" << endl;
        out << " $PDC   VDWINC=" << doubleSpinBox_VDWINC->value ()
            << "        $END" << endl;
        out << " $PDC   LAYER="  << spinBox_LAYER->value ()
            << "          $END" << endl;
        out << " $ELMOM IEMOM="  << spinBox_IEMOM->value ()
            << "          $END" << endl;

        out << " $ELMOM  WHERE=POINTS    $END" << endl;
        out << " $POINTS" << endl;
        out << " ANGS 1" << endl;
        out << " 0 0 0" << endl;
        out << " $END" << endl;

        out << " $MP2   MP2PRP=.true.    $END" << endl; 
        out << endl;
    }

    {//fix
        out << endl;
        out << "# $p2p p2p=.t. dlb=.t. xdlb=.t.           $END" << endl;
        out << "# $MP2 METHOD=1                           $END" << endl;
        //out << "# $MP2GRD TOL1=1D-15 TOL2=1D-15           $END" << endl;
        out << "# $p2p mxbuf=2048 $end                    $END" << endl;
        out << "# $smp csmtx=.t. $end                     $END" << endl;
        out << endl;

        out << "# $CONTRL nzvar=1                         $END" << endl;
        out << "# $zmat dlc=1 auto=1 nonvdw(1)=3,6        $END" << endl;
        out << endl;

    }

    out << data().c_str() << endl;
}

Text Panel_GAMESS::
basis ()
{
    external_basis_.clear();

    Text text, panel = tabWidget_Basis->currentWidget()->objectName();

    if (panel == "Standard")
    {
        if (listWidget_Basis->currentItem()->text() == "MINI")
        {
            text += 
                " $BASIS GBASIS=MINI           $END\n";
        }
        if (listWidget_Basis->currentItem()->text() == "3-21+G")
        {
            text += 
                " $BASIS GBASIS=N21            $END\n"
                " $BASIS NGAUSS=3              $END\n"
                " $BASIS diffSP=.true.         $END\n";
        }
        if (listWidget_Basis->currentItem()->text() == "6-31+G(d)")
        {
            text += 
                " $BASIS GBASIS=N31            $END\n"
                " $BASIS NGAUSS=6              $END\n"
                " $BASIS nDfunc=1              $END\n"
                " $BASIS diffSP=.true.         $END\n";
        }
        if (listWidget_Basis->currentItem()->text() == "6-311+G(2d,p)")
        {
            text += 
                " $BASIS GBASIS=N311           $END\n"
                " $BASIS NGAUSS=6              $END\n"
                " $BASIS nDfunc=2              $END\n"
                " $BASIS nPfunc=1              $END\n"
                " $BASIS diffSP=.true.         $END\n";
        }
        if (listWidget_Basis->currentItem()->text() == "6-311+G(3df,2p)")
        {
            text += 
                " $BASIS GBASIS=N311           $END\n"
                " $BASIS NGAUSS=6              $END\n"
                " $BASIS nDfunc=3              $END\n"
                " $BASIS nFfunc=1              $END\n"
                " $BASIS nPfunc=2              $END\n"
                " $BASIS diffSP=.true.         $END\n";
        }
        if (listWidget_Basis->currentItem()->text() == "aug-cc-pVDZ")
        {
            text += " $BASIS extfil=.t. gbasis=acc-pvdz  $END\n";
            external_basis_ = "acc-pvdz";
        }
        if (listWidget_Basis->currentItem()->text() == "aug-cc-pVTZ")
        {
            text += " $BASIS extfil=.t. gbasis=acc-pvtz  $END\n";
            external_basis_ = "acc-pvtz";
        }
        if (listWidget_Basis->currentItem()->text() == "aug-cc-pVQZ")
        {
            text += " $BASIS extfil=.t. gbasis=acc-pvqz  $END\n";
            external_basis_ = "acc-pvqz";
        }
        if (listWidget_Basis->currentItem()->text() == "cc-pVDZ")
        {
            text += " $BASIS extfil=.t. gbasis=cc-pvdz  $END\n";
            external_basis_ = "cc-pvdz";
        }
        if (listWidget_Basis->currentItem()->text() == "cc-pVTZ")
        {
            text += " $BASIS extfil=.t. gbasis=cc-pvtz  $END\n";
            external_basis_ = "cc-pvtz";
        }
        if (listWidget_Basis->currentItem()->text() == "cc-pVQZ")
        {
            text += " $BASIS extfil=.t. gbasis=cc-pvqz  $END\n";
            external_basis_ = "cc-pvqz";
        }      
        
        if (listWidget_Basis->currentItem()->text() == "sbsse-1")
        {
            text += " $BASIS extfil=.t. gbasis=sbsse-1  $END\n";
            external_basis_ = "sbsse-1";
        }      
        if (listWidget_Basis->currentItem()->text() == "sbsse-2")
        {
            text += " $BASIS extfil=.t. gbasis=sbsse-2  $END\n";
            external_basis_ = "sbsse-2";
        }      
        
        
        if (listWidget_Basis->currentItem()->text() == "user")
        {
            text += " $BASIS extfil=.t. gbasis=user  $END\n";
            external_basis_ = "user";
        }

        if (listWidget_Basis->currentItem()->text() == "")
        {
        }
    }
    else // Assembly
    {
        Text basis = comboBox_Basis->currentText();

        if (basis != "N311" && basis != "N31" && 
            basis != "N21"  && basis != "STO")
        {
            text += " $BASIS GBASIS=";
            text += basis;
            text += "           $END\n";
            return text;
        }

        text += " $BASIS NGAUSS=";
        text += spinBox_Gauss->value();
        text += "              $END\n";

        text += " $BASIS GBASIS=";
        text += basis;
        text += "           $END\n";

        if (checkBox_diffSP->isChecked())
            text += " $BASIS diffSP=.true.         $END\n";

        if (checkBox_diffS->isChecked())
            text += " $BASIS diffS=.true.          $END\n";

        text += " $BASIS nDfunc=";
        text += spinBox_D->value();
        text += "              $END\n";

        if (checkBox_F->isChecked())
            text += " $BASIS nFfunc=1              $END\n";

        text += " $BASIS nPfunc=";
        text += spinBox_P->value();
        text += "              $END\n";

        if (checkBox_SPLIT3->isChecked())
        {
            text += " $BASIS SPLIT3(1)=";
            text += lineEdit_SPLIT3->text();
            text += "  $END\n";
        }

        if (checkBox_SPLIT2->isChecked())
        {
            text += " $BASIS SPLIT2(1)=";
            text += lineEdit_SPLIT2->text();
            text += "  $END\n";
        }
    }

    return text;
}

Text Panel_GAMESS::
data ()
{
    Text text( 
        " $DATA\n"
        "-----------------------------------------------------\n"
        "C1\n");

    for (int i=0;  i<model_->atom_count();  ++i)
    {
        Atom const & atom = model_->atom(i);

        text += " ";
        text += atom.element().c_str();
        text += "    ";
        text += atom.element().number();
        text += "    ";
        text += Text(atom.x(), 15);
        text += " ";
        text += Text(atom.y(), 15);
        text += " ";
        text += Text(atom.z(), 15);
        text += "\n";
    }
    text += " $END\n";
    return text;
}

bool Panel_GAMESS::
not_semiempirical () const
{
    Text basis = comboBox_Basis->currentText();
    Text panel = tabWidget_Basis->currentWidget()->objectName();

    if (panel == "Assembly" &&
        (basis == "MNDO" || basis == "AM1" || basis == "PM3" || basis == "RM1"))
        return false;

    return true;
}

bool Panel_GAMESS::
is_spherical_basis_set () const
{
    Text basis = listWidget_Basis->currentItem()->text();
    Text panel = tabWidget_Basis->currentWidget()->objectName();

    if (panel == "Standard" &&
        (basis == "aug-cc-pVDZ" || 
         basis == "aug-cc-pVTZ" || 
         basis == "aug-cc-pVQZ" || 
         basis == "cc-pVDZ" || 
         basis == "cc-pVTZ" || 
         basis == "cc-pVQZ" || 
         basis == "sbsse-2"
         ))
        return true;

    return false;
}

}//MM
