#include "Model_adapter.h"

#include "Model.h"
#include "Model_kit.h"
#include "Atom.h"
#include "Atom_kit.h"

#include <QApplication>//fix

namespace MM
{

int Model_adapter::
size () const   
{
    return 3 * model_.atom_count ();
}

void Model_adapter::
add (int i, real value)
{
    int n = i/3;
    Atom & atom = model_.atom (n);
    
    if (atom.kit().was_frozen())
        return;

    if (!coords_was_changed_[n])
    {
        change_list_.push_back (n);
        coords_was_changed_ [n] = true;
    }

    //switch (i%3)
    //{
    //    case 0: atom.set_x (value + atom.x());  break;
    //    case 1: atom.set_y (value + atom.y());  break;
    //    case 2: atom.set_z (value + atom.z());  break;
    //}
    switch (i%3)
    {
        case 0: atom.move (Vector_3D_impl(value,0,0), true);  break;
        case 1: atom.move (Vector_3D_impl(0,value,0), true);  break;
        case 2: atom.move (Vector_3D_impl(0,0,value), true);  break;
    }

    E_flag_ = 0;
}

real Model_adapter::
get (int i) const
{
    switch (i%3)
    {
        //fix x___;
        case 0: return model_.atom (i/3).x();
        case 1: return model_.atom (i/3).y();
        case 2: return model_.atom (i/3).z();
    }
    return 0;
}

void Model_adapter::
reject_changes()
{
    E_           = previous_E_;
    E_flag_      = previous_E_flag_;

    int i=0, n;

    for (;  i<change_list_.size();  i++)
    {
        n = change_list_[i];
        Atom & atom = model_.atom (n);

        //model_.atom (n). set_x (previous_[n].x());
        //model_.atom (n). set_y (previous_[n].y());
        //model_.atom (n). set_z (previous_[n].z());

        atom.set_position (Point_3D_impl (previous_[n].x(), 
                                          previous_[n].y(), 
                                          previous_[n].z()), true);

        coords_was_changed_ [n] = false;
    }
    change_list_.clear();
}

void Model_adapter::
accept_changes()
{
    previous_E_            = E_;
    previous_E_flag_       = E_flag_;

    int i=0, n;

    for (;  i<change_list_.size();  i++)
    {
        n = change_list_[i];
        previous_[n].set_x (model_.atom(n).x());
        previous_[n].set_y (model_.atom(n).y());
        previous_[n].set_z (model_.atom(n).z());

        coords_was_changed_ [n] = false;
    }
    change_list_.clear ();

    model_.kit().event().conformation_changed ();
    model_.kit().visual().renew ();
    qApp->processEvents(QEventLoop::ExcludeSocketNotifiers);
}

void Model_adapter::
grad (Array <real> & gr) 
{
    int coord_count = size();
    int atom_count  = model_.atom_count();

    if (gr.size() != coord_count)
    {
        gr.clear();
        gr.reserve (coord_count);
    }

    bool ad_hoc = model_.kit().interaction().common().ad_hoc();
    model_.kit().force_field().start_work (!ad_hoc);
    {
        model_.kit().interaction()./*add_*/force();
    }
    if  (!ad_hoc && model_.kit().force_field().was_problems ())
        throw Text ("Force Field problem");

    for (int i=0;  i<atom_count;  ++i)
    {
        Vector_3D const & force = model_.atom(i).kit().potential_force ();
        gr [3*i  ]  = -force.x();
        gr [3*i+1]  = -force.y();
        gr [3*i+2]  = -force.z();
    }
}

real Model_adapter::
value() 
{
    //FIX; 
    //return model_.kit().interaction().potential ();

    if (E_flag_)
        return E_;

    E_flag_ = true;
    
    bool ad_hoc = model_.kit().interaction().common().ad_hoc();
    model_.kit().force_field().start_work (!ad_hoc);
    {
        E_ = model_.kit().interaction().potential ();
    }
    if  (!ad_hoc && model_.kit().force_field().was_problems ())
        throw Text ("Force Field problem");

    return E_;
}

Model_adapter::
Model_adapter (Model & model) 
    : 
    model_(model), current_(0),
    E_              (0.),            
    E_flag_         (0),
    previous_E_     (0.),   
    previous_E_flag_(0)
{
    model.kit().strip (true);

    int i;
    change_list_.reserve (model.atom_count());

    for (i=0;  i<model.atom_count();  ++i)
    {
        previous_.push_back (Point_3D_impl (0,0,0));
        coords_was_changed_.push_back (false);
    }

    for (i=0;  i<previous_.size();  ++i)
    {
        previous_[i].set_x (model_.atom(i).x());
        previous_[i].set_y (model_.atom(i).y());
        previous_[i].set_z (model_.atom(i).z());
    }
}

}//MM
